@@ -36,7 +36,7 @@ namespace Mono.Debugger.Soft
 
 	struct SourceInfo {
 		public string source_file;
-		public byte[] guid, hash;
+		public byte[] hash;
 	}
 
 	class DebugInfo {
@@ -420,7 +420,7 @@ namespace Mono.Debugger.Soft
 		 * with newer runtimes, and vice versa.
 		 */
 		internal const int MAJOR_VERSION = 2;
-		internal const int MINOR_VERSION = 45;
+		internal const int MINOR_VERSION = 48;
 
 		enum WPSuspendPolicy {
 			NONE = 0,
@@ -442,7 +442,8 @@ namespace Mono.Debugger.Soft
 			TYPE = 23,
 			MODULE = 24,
 			FIELD = 25,
-			EVENT = 64
+			EVENT = 64,
+			POINTER = 65
 		}
 
 		enum EventKind {
@@ -533,7 +534,12 @@ namespace Mono.Debugger.Soft
 			GET_OBJECT = 4,
 			GET_TYPE = 5,
 			GET_NAME = 6,
-			GET_DOMAIN = 7
+			GET_DOMAIN = 7,
+			GET_METADATA_BLOB = 8,
+			GET_IS_DYNAMIC = 9,
+			GET_PDB_BLOB = 10,
+			GET_TYPE_FROM_TOKEN = 11,
+			GET_METHOD_FROM_TOKEN = 12
 		}
 
 		enum CmdModule {
@@ -574,7 +580,8 @@ namespace Mono.Debugger.Soft
 			GET_INTERFACES = 16,
 			GET_INTERFACE_MAP = 17,
 			IS_INITIALIZED = 18,
-			CREATE_INSTANCE = 19
+			CREATE_INSTANCE = 19,
+			GET_VALUE_SIZE = 20
 		}
 
 		enum CmdField {
@@ -586,6 +593,11 @@ namespace Mono.Debugger.Soft
 			BINDING_FLAGS_IGNORE_CASE = 0x70000000,
 		}
 
+		enum MemberListTypeExtensions {
+			CaseSensitive = 1,
+			CaseInsensitive = 2
+		}
+
 		enum CmdStackFrame {
 			GET_VALUES = 1,
 			GET_THIS = 2,
@@ -606,6 +618,10 @@ namespace Mono.Debugger.Soft
 			GET_CHARS = 3
 		}
 
+		enum CmdPointer {
+			GET_VALUE = 1
+		}
+
 		enum CmdObjectRef {
 			GET_TYPE = 1,
 			GET_VALUES = 2,
@@ -642,6 +658,17 @@ namespace Mono.Debugger.Soft
 			return packet [offset++];
 		}
 
+		static byte[] decode_bytes (byte[] packet, ref int offset, int length)
+		{
+			if (length + offset > packet.Length)
+				throw new ArgumentOutOfRangeException ();
+
+			var bytes = new byte[length];
+			Array.Copy (packet, offset, bytes, 0, length);
+			offset += length;
+			return bytes;
+		}
+
 		static int decode_short (byte[] packet, ref int offset) {
 			int res = ((int)packet [offset] << 8) | (int)packet [offset + 1];
 			offset += 2;
@@ -730,10 +757,12 @@ namespace Mono.Debugger.Soft
 		}
 
 		class PacketReader {
+			Connection connection;
 			byte[] packet;
 			int offset;
 
-			public PacketReader (byte[] packet) {
+			public PacketReader (Connection connection, byte[] packet) {
+				this.connection = connection;
 				this.packet = packet;
 
 				// For event packets
@@ -845,9 +874,16 @@ namespace Mono.Debugger.Soft
 					return new ValueImpl { Type = etype, Value = ReadDouble () };
 				case ElementType.I:
 				case ElementType.U:
-				case ElementType.Ptr:
 					// FIXME: The client and the debuggee might have different word sizes
 					return new ValueImpl { Type = etype, Value = ReadLong () };
+				case ElementType.Ptr:
+					long value = ReadLong ();
+					if (connection.Version.AtLeast (2, 46)) {
+						long pointerClass = ReadId ();
+						return new ValueImpl { Type = etype, Klass = pointerClass, Value = value };
+					} else {
+						return new ValueImpl { Type = etype, Value = value };
+					}
 				case ElementType.String:
 				case ElementType.SzArray:
 				case ElementType.Class:
@@ -880,6 +916,15 @@ namespace Mono.Debugger.Soft
 					res [i] = ReadId ();
 				return res;
 			}
+			
+			public byte[] ReadByteArray () {
+				var length = ReadInt ();
+				return decode_bytes (packet, ref offset, length);
+			}
+
+			public bool ReadBool () {
+				return ReadByte () != 0;
+			}
 		}
 
 		class PacketWriter {
@@ -1236,6 +1281,8 @@ namespace Mono.Debugger.Soft
 					bool res = ReceivePacket ();
 					if (!res)
 						break;
+				} catch (ThreadAbortException) {
+					break;
 				} catch (Exception ex) {
 					if (!closed) {
 						Console.WriteLine (ex);
@@ -1281,7 +1328,7 @@ namespace Mono.Debugger.Soft
 					if (cb != null)
 						cb.Invoke (id, packet);
 				} else {
-					PacketReader r = new PacketReader (packet);
+					PacketReader r = new PacketReader (this, packet);
 
 					if (r.CommandSet == CommandSet.EVENT && r.Command == (int)CmdEvent.COMPOSITE) {
 						int spolicy = r.ReadByte ();
@@ -1500,7 +1547,7 @@ namespace Mono.Debugger.Soft
 						if (EnableConnectionLogging)
 							LogPacket (packet_id, encoded_packet, p, command_set, command, watch);
 						/* Run the callback on a tp thread to avoid blocking the receive thread */
-						PacketReader r = new PacketReader (p);
+						PacketReader r = new PacketReader (this, p);
 						cb.BeginInvoke (r, null, null);
 					};
 					reply_cb_counts [id] = count;
@@ -1547,7 +1594,7 @@ namespace Mono.Debugger.Soft
 					if (reply_packets.ContainsKey (packetId)) {
 						byte[] reply = reply_packets [packetId];
 						reply_packets.Remove (packetId);
-						PacketReader r = new PacketReader (reply);
+						PacketReader r = new PacketReader (this, reply);
 
 						if (EnableConnectionLogging)
 							LogPacket (packetId, encoded_packet, reply, command_set, command, watch);
@@ -2118,6 +2165,26 @@ namespace Mono.Debugger.Soft
 		internal long Assembly_GetIdDomain (long id) {
 			return SendReceive (CommandSet.ASSEMBLY, (int)CmdAssembly.GET_DOMAIN, new PacketWriter ().WriteId (id)).ReadId ();
 		}
+		
+		internal byte[] Assembly_GetMetadataBlob (long id) {
+			return SendReceive (CommandSet.ASSEMBLY, (int)CmdAssembly.GET_METADATA_BLOB, new PacketWriter ().WriteId (id)).ReadByteArray ();
+		}
+
+		internal bool Assembly_IsDynamic (long id) {
+			return SendReceive (CommandSet.ASSEMBLY, (int) CmdAssembly.GET_IS_DYNAMIC, new PacketWriter ().WriteId (id)).ReadBool ();
+		}
+		
+		internal byte[] Assembly_GetPdbBlob (long id) {
+			return SendReceive (CommandSet.ASSEMBLY, (int)CmdAssembly.GET_PDB_BLOB, new PacketWriter ().WriteId (id)).ReadByteArray ();
+		}
+
+		internal long Assembly_GetType (long id, uint token) {
+			return SendReceive (CommandSet.ASSEMBLY, (int)CmdAssembly.GET_TYPE_FROM_TOKEN, new PacketWriter ().WriteId (id).WriteInt ((int)token)).ReadId ();
+		}
+
+		internal long Assembly_GetMethod (long id, uint token) {
+			return SendReceive (CommandSet.ASSEMBLY, (int)CmdAssembly.GET_METHOD_FROM_TOKEN, new PacketWriter ().WriteId (id).WriteInt ((int)token)).ReadId ();
+		}
 
 		/*
 		 * TYPE
@@ -2212,7 +2279,7 @@ namespace Mono.Debugger.Soft
 		internal ValueImpl[] Type_GetValues (long id, long[] fields, long thread_id) {
 			int len = fields.Length;
 			PacketReader r;
-			if (thread_id != 0)
+			if (thread_id != 0 && Version.AtLeast(2, 3))
 				r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUES_2, new PacketWriter ().WriteId (id).WriteId (thread_id).WriteInt (len).WriteIds (fields));
 			else
 				r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUES, new PacketWriter ().WriteId (id).WriteInt (len).WriteIds (fields));
@@ -2257,7 +2324,11 @@ namespace Mono.Debugger.Soft
 
 		public long[] Type_GetMethodsByNameFlags (long id, string name, int flags, bool ignoreCase) {
 			flags |= ignoreCase ? (int)BindingFlagsExtensions.BINDING_FLAGS_IGNORE_CASE : 0;
-			PacketReader r = SendReceive (CommandSet.TYPE, (int)CmdType.CMD_TYPE_GET_METHODS_BY_NAME_FLAGS, new PacketWriter ().WriteId (id).WriteString (name).WriteInt (flags));
+			int listType = ignoreCase ? (int)MemberListTypeExtensions.CaseInsensitive : (int)MemberListTypeExtensions.CaseSensitive;
+			var w = new PacketWriter ().WriteId (id).WriteString (name).WriteInt (flags);
+			if (Version.AtLeast (2, 48))
+				w.WriteInt (listType);
+			PacketReader r = SendReceive (CommandSet.TYPE, (int)CmdType.CMD_TYPE_GET_METHODS_BY_NAME_FLAGS, w);
 			int len = r.ReadInt ();
 			long[] res = new long [len];
 			for (int i = 0; i < len; ++i)
@@ -2295,6 +2366,11 @@ namespace Mono.Debugger.Soft
 			return r.ReadId ();
 		}
 
+		internal int Type_GetValueSize (long id) {
+			PacketReader r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUE_SIZE, new PacketWriter ().WriteId (id));
+			return r.ReadInt ();
+		}
+
 		/*
 		 * FIELD
 		 */
@@ -2473,7 +2549,16 @@ namespace Mono.Debugger.Soft
 			for (int i = 0; i < length; ++i)
 				res [i] = (char)r.ReadShort ();
 			return res;
-		}			
+		}
+
+		/*
+		 * POINTERS
+		 */
+
+		internal ValueImpl Pointer_GetValue (long address, TypeMirror type)
+		{
+			return SendReceive (CommandSet.POINTER, (int)CmdPointer.GET_VALUE, new PacketWriter ().WriteLong (address).WriteId (type.Id)).ReadValue ();
+		}
 
 		/*
 		 * OBJECTS
