@@ -38,6 +38,7 @@ using System.Security;
 using System.Reflection;
 using System.Threading;
 using System.Runtime.InteropServices.ComTypes;
+using System.Text;
 
 using System.Runtime.ConstrainedExecution;
 #if !FULL_AOT_RUNTIME
@@ -83,6 +84,9 @@ namespace System.Runtime.InteropServices
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static IntPtr AllocCoTaskMem (int cb);
+		
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		internal extern static IntPtr AllocCoTaskMemSize (UIntPtr sizet);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		[ReliabilityContractAttribute (Consistency.WillNotCorruptState, Cer.MayFail)]
@@ -106,92 +110,176 @@ namespace System.Runtime.InteropServices
 			throw new NotImplementedException ();
 		}
 
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		internal extern static void copy_to_unmanaged (Array source, int startIndex,
-							       IntPtr destination, int length);
+		unsafe internal static void copy_to_unmanaged (Array source, int startIndex,
+		       IntPtr destination, int length)
+		{
+			copy_to_unmanaged_fixed (source, startIndex, destination, length, null);
+		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		internal extern static void copy_from_unmanaged (IntPtr source, int startIndex,
-								 Array destination, int length);
+		unsafe private extern static void copy_to_unmanaged_fixed (Array source, int startIndex,
+		       IntPtr destination, int length, void* fixed_source_element);
 
-		public static void Copy (byte[] source, int startIndex, IntPtr destination, int length)
+		static private bool skip_fixed (System.Array array, int startIndex)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			// In particular, we see length == 0 && startIndex == array.Length, and fixed fails.
+			return startIndex < 0 || startIndex >= array.Length;
 		}
 
-		public static void Copy (char[] source, int startIndex, IntPtr destination, int length)
+		unsafe internal static void copy_to_unmanaged (byte[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			// This function is inconsistent with its surroundings.
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, null);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (short[] source, int startIndex, IntPtr destination, int length)
+		unsafe internal static void copy_to_unmanaged (char[] source, int startIndex,
+		       IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			// This function is inconsistent with its surroundings.
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, null);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (int[] source, int startIndex, IntPtr destination, int length)
+		public unsafe static void Copy (byte[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (long[] source, int startIndex, IntPtr destination, int length)
+		public unsafe static void Copy (char[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (float[] source, int startIndex, IntPtr destination, int length)
+		public unsafe static void Copy (short[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (double[] source, int startIndex, IntPtr destination, int length)
+		public unsafe static void Copy (int[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (IntPtr[] source, int startIndex, IntPtr destination, int length)
+		public unsafe static void Copy (long[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_to_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (IntPtr source, byte[] destination, int startIndex, int length)
+		public unsafe static void Copy (float[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (IntPtr source, char[] destination, int startIndex, int length)
+		public unsafe static void Copy (double[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (IntPtr source, short[] destination, int startIndex, int length)
+		public unsafe static void Copy (IntPtr[] source, int startIndex, IntPtr destination, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (source, startIndex))
+				copy_to_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_source = &source [startIndex])
+				copy_to_unmanaged_fixed (source, startIndex, destination, length, fixed_source);
 		}
 
-		public static void Copy (IntPtr source, int[] destination, int startIndex, int length)
+		unsafe internal static void copy_from_unmanaged (IntPtr source, int startIndex, Array destination, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			copy_from_unmanaged_fixed (source, startIndex, destination, length, null);
 		}
 
-		public static void Copy (IntPtr source, long[] destination, int startIndex, int length)
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		unsafe private extern static void copy_from_unmanaged_fixed (IntPtr source, int startIndex,
+			Array destination, int length, void* fixed_destination_element);
+
+		public unsafe static void Copy (IntPtr source, byte[] destination, int startIndex, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
 		}
 
-		public static void Copy (IntPtr source, float[] destination, int startIndex, int length)
+		public unsafe static void Copy (IntPtr source, char[] destination, int startIndex, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
 		}
 
-		public static void Copy (IntPtr source, double[] destination, int startIndex, int length)
+		public unsafe static void Copy (IntPtr source, short[] destination, int startIndex, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
 		}
 
-		public static void Copy (IntPtr source, IntPtr[] destination, int startIndex, int length)
+		public unsafe static void Copy (IntPtr source, int[] destination, int startIndex, int length)
 		{
-			copy_from_unmanaged (source, startIndex, destination, length);
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
+		}
+
+		public unsafe static void Copy (IntPtr source, long[] destination, int startIndex, int length)
+		{
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
+		}
+
+		public unsafe static void Copy (IntPtr source, float[] destination, int startIndex, int length)
+		{
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
+		}
+
+		public unsafe static void Copy (IntPtr source, double[] destination, int startIndex, int length)
+		{
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
+		}
+
+		public unsafe static void Copy (IntPtr source, IntPtr[] destination, int startIndex, int length)
+		{
+			if (skip_fixed (destination, startIndex))
+				copy_from_unmanaged (source, startIndex, destination, length);
+			else fixed (void* fixed_destination = &destination [startIndex])
+				copy_from_unmanaged_fixed (source, startIndex, destination, length, fixed_destination);
 		}
 
 		public static IntPtr CreateAggregatedObject (IntPtr pOuter,
@@ -285,6 +373,12 @@ namespace System.Runtime.InteropServices
 			FreeCoTaskMem (s);
 		}
 
+		public static void ZeroFreeCoTaskMemUTF8 (IntPtr s)
+		{
+			ClearAnsi (s);
+			FreeCoTaskMem (s);
+		}
+		
 		public static void ZeroFreeGlobalAllocAnsi (IntPtr s)
 		{
 			ClearAnsi (s);
@@ -731,7 +825,7 @@ namespace System.Runtime.InteropServices
 #else
 		public static bool IsComObject (object o)
 		{
-			throw new PlatformNotSupportedException ();
+			return false;
 		}
 #endif
 
@@ -758,6 +852,16 @@ namespace System.Runtime.InteropServices
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static string PtrToStringAnsi (IntPtr ptr, int len);
 
+		public static string PtrToStringUTF8 (IntPtr ptr)
+		{
+			return PtrToStringAnsi (ptr);
+		}
+		
+		public static string PtrToStringUTF8 (IntPtr ptr, int byteLen)
+		{
+			return PtrToStringAnsi (ptr, byteLen);
+		}
+		
 		public static string PtrToStringAuto (IntPtr ptr)
 		{
 			return SystemDefaultCharSize == 2
@@ -1050,26 +1154,17 @@ namespace System.Runtime.InteropServices
 			return (size + 3) & (~((uint)3));
 		}
 
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public extern static IntPtr StringToBSTR (string s);
+		public unsafe static IntPtr StringToBSTR (string s)
+		{
+			if (s == null)
+				return IntPtr.Zero;
+			fixed (char* fixed_s = s)
+				return BufferToBSTR (fixed_s, s.Length);
+		}
 
-		//
-		// I believe this is wrong, because in Mono and in P/Invoke
-		// we treat "Ansi" conversions as UTF-8 conversions, while
-		// this one does not do this
-		//
 		public static IntPtr StringToCoTaskMemAnsi (string s)
 		{
-			int length = s.Length + 1;
-			IntPtr ctm = AllocCoTaskMem (length);
-
-			byte[] asBytes = new byte[length];
-			for (int i = 0; i < s.Length; i++)
-				asBytes[i] = (byte)s[i];
-			asBytes[s.Length] = 0;
-
-			copy_to_unmanaged (asBytes, 0, ctm, length);
-			return ctm;
+			return StringToAllocatedMemoryUTF8 (s);
 		}
 
 		public static IntPtr StringToCoTaskMemAuto (string s)
@@ -1092,8 +1187,41 @@ namespace System.Runtime.InteropServices
 		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public extern static IntPtr StringToHGlobalAnsi (string s);
+		unsafe extern static IntPtr StringToHGlobalAnsi (char* s, int length);
 
+		public unsafe static IntPtr StringToHGlobalAnsi (string s)
+		{
+			fixed (char* fixed_s = s)
+				return StringToHGlobalAnsi (fixed_s, (s != null) ? s.Length : 0);
+		}
+
+		unsafe public static IntPtr StringToAllocatedMemoryUTF8(String s)
+		{
+			const int MAX_UTF8_CHAR_SIZE = 3;
+			if (s == null)
+				return IntPtr.Zero;
+
+			int nb = (s.Length + 1) * MAX_UTF8_CHAR_SIZE;
+
+			// Overflow checking
+			if (nb < s.Length)
+				throw new ArgumentOutOfRangeException("s");
+			
+			IntPtr pMem = AllocCoTaskMemSize(new UIntPtr((uint)nb +1));
+			
+			if (pMem == IntPtr.Zero)
+				throw new OutOfMemoryException();
+
+			byte* pbMem = (byte*)pMem;
+
+            fixed (char* pwzChar = s)
+            {
+                int nbWritten = Encoding.UTF8.GetBytes(pwzChar, s.Length, pbMem, nb);
+				pbMem[nbWritten] = 0;
+            }
+			return pMem;
+		}
+		
 		public static IntPtr StringToHGlobalAuto (string s)
 		{
 			return SystemDefaultCharSize == 2
@@ -1101,9 +1229,15 @@ namespace System.Runtime.InteropServices
 		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public extern static IntPtr StringToHGlobalUni (string s);
+		unsafe extern static IntPtr StringToHGlobalUni (char* s, int length);
 
-		public static IntPtr SecureStringToBSTR (SecureString s)
+		public unsafe static IntPtr StringToHGlobalUni (string s)
+		{
+			fixed (char* fixed_s = s)
+				return StringToHGlobalUni (fixed_s, (s != null) ? s.Length : 0);
+		}
+
+		public unsafe static IntPtr SecureStringToBSTR (SecureString s)
 		{
 			if (s == null)
 				throw new ArgumentNullException ("s");
@@ -1120,8 +1254,9 @@ namespace System.Runtime.InteropServices
 					buffer[i + 1] = b;
 				}
 			}
-			return BufferToBSTR (buffer, len);
-        }
+			fixed (byte* fixed_buffer = buffer)
+				return BufferToBSTR ((char*)fixed_buffer, len);
+		}
 
 		public static IntPtr SecureStringToCoTaskMemAnsi (SecureString s)
 		{
@@ -1210,7 +1345,7 @@ namespace System.Runtime.InteropServices
 
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public extern static IntPtr BufferToBSTR (Array ptr, int slen);
+		extern unsafe static IntPtr BufferToBSTR (char* ptr, int slen);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public extern static IntPtr UnsafeAddrOfPinnedArrayElement (Array arr, int index);
@@ -1426,10 +1561,10 @@ namespace System.Runtime.InteropServices
 			const int COR_E_TARGET = unchecked ((int)0x80131603L);
 			const int COR_E_TARGETINVOCATION = unchecked ((int)0x80131604L);
 			const int COR_E_TARGETPARAMCOUNT = unchecked ((int)0x8002000EL);
-			const int COR_E_THREADABORTED = unchecked ((int)0x80131530L);
+			//const int COR_E_THREADABORTED = unchecked ((int)0x80131530L);
 			const int COR_E_THREADINTERRUPTED = unchecked ((int)0x80131519L);
 			const int COR_E_THREADSTATE = unchecked ((int)0x80131520L);
-			const int COR_E_THREADSTOP = unchecked ((int)0x80131521L);
+			//const int COR_E_THREADSTOP = unchecked ((int)0x80131521L);
 			const int COR_E_TYPEINITIALIZATION = unchecked ((int)0x80131534L);
 			const int COR_E_VERIFICATION = unchecked ((int)0x8013150DL);
 			//const int COR_E_WEAKREFERENCE = unchecked ((int)?);
@@ -1722,5 +1857,112 @@ namespace System.Runtime.InteropServices
 		internal static void SetLastWin32Error (int error)
 		{
 		}
+
+#if FEATURE_COMINTEROP || MONO_COM
+		// Copied from referencesource/mscorlib/system/runtime/interopservices/marshal.cs
+		//====================================================================
+		// return the raw IUnknown* for a COM Object not related to current 
+		// context
+		// Does not call AddRef
+		//====================================================================
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		internal static extern IntPtr /* IUnknown* */ GetRawIUnknownForComObjectNoAddRef(Object o);
+		
+		// Copied from referencesource/mscorlib/system/runtime/interopservices/marshal.cs
+		//====================================================================
+		// Converts the CLR exception to an HRESULT. This function also sets
+		// up an IErrorInfo for the exception.
+		// This function is only used in WinRT and converts ObjectDisposedException
+		// to RO_E_CLOSED
+		//====================================================================
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		internal static extern int GetHRForException_WinRT(Exception e);
+
+		// Copied from referencesource/mscorlib/system/runtime/interopservices/marshal.cs
+		//========================================================================
+		// Create activation factory and wraps it with a unique RCW
+		//========================================================================
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		internal static extern object GetNativeActivationFactory(Type type);
+#else
+		internal static IntPtr /* IUnknown* */ GetRawIUnknownForComObjectNoAddRef(Object o) {
+			throw new NotSupportedException();
+		}
+
+		internal static int GetHRForException_WinRT(Exception e) {
+			throw new NotSupportedException();
+		}
+
+		internal static object GetNativeActivationFactory(Type type) {
+			throw new NotSupportedException();
+		}
+#endif
+
+		internal class MarshalerInstanceKeyComparer : IEqualityComparer<(Type, string)> {
+			public bool Equals ((Type, string) lhs, (Type, string) rhs) {
+				return lhs.CompareTo(rhs) == 0;
+			}
+
+			public int GetHashCode ((Type, string) key) {
+				return key.GetHashCode ();
+			}
+		}
+
+		internal static Dictionary<(Type, string), ICustomMarshaler> MarshalerInstanceCache;
+		internal static readonly object MarshalerInstanceCacheLock = new object ();
+
+		internal static ICustomMarshaler GetCustomMarshalerInstance (Type type, string cookie) {
+			var key = (type, cookie);
+
+			LazyInitializer.EnsureInitialized (
+				ref MarshalerInstanceCache, 
+				() => new Dictionary<(Type, string), ICustomMarshaler> (new MarshalerInstanceKeyComparer ())
+			);
+
+			ICustomMarshaler result;
+			bool gotExistingInstance;
+			lock (MarshalerInstanceCacheLock)
+				gotExistingInstance = MarshalerInstanceCache.TryGetValue (key, out result);
+
+			if (!gotExistingInstance) {
+				MonoMethod getInstanceMethod;
+				try {
+					getInstanceMethod = (MonoMethod)type.GetMethod (
+						"GetInstance", BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.InvokeMethod,
+						null, new Type[] { typeof(string) }, null
+					);
+				} catch (AmbiguousMatchException) {
+					throw new ApplicationException ($"Custom marshaler '{type.FullName}' implements multiple static GetInstance methods that take a single string parameter.");
+				}
+
+				if ((getInstanceMethod == null) || 
+					(getInstanceMethod.ReturnType != typeof (ICustomMarshaler))) {
+					throw new ApplicationException ($"Custom marshaler '{type.FullName}' does not implement a static GetInstance method that takes a single string parameter and returns an ICustomMarshaler.");
+				}
+
+				Exception exc;
+				try {
+					result = (ICustomMarshaler)getInstanceMethod.InternalInvoke (null, new object[] { cookie }, out exc);
+				} catch (Exception e) {
+					// FIXME: mscorlib's legacyUnhandledExceptionPolicy is apparently 1, 
+					//  so exceptions are thrown instead of being passed through the outparam
+					exc = e;
+					result = null;
+				}
+
+				if (exc != null) {
+					var edi = System.Runtime.ExceptionServices.ExceptionDispatchInfo.Capture (exc);
+					edi.Throw ();
+				}
+
+				if (result == null)
+					throw new ApplicationException ($"A call to GetInstance() for custom marshaler '{type.FullName}' returned null, which is not allowed.");
+
+				lock (MarshalerInstanceCacheLock)
+					MarshalerInstanceCache[key] = result;
+			}
+
+			return result;
+		}
 	}
 }
