@@ -62,6 +62,8 @@ namespace System
 		private IntPtr delegate_trampoline;
 		private IntPtr extra_arg;
 		private IntPtr method_code;
+		private IntPtr interp_method;
+		private IntPtr interp_invoke_impl;
 		private MethodInfo method_info;
 
 		// Keep a ref of the MethodInfo passed to CreateDelegate.
@@ -115,6 +117,8 @@ namespace System
 			}
 		}
 
+		internal IntPtr GetNativeFunctionPointer () => method_ptr;
+
 		//
 		// Methods
 		//
@@ -134,6 +138,8 @@ namespace System
 			if (!match) {
 				if (delArgType.IsEnum && Enum.GetUnderlyingType (delArgType) == argType)
 					match = true;
+				else if (argType.IsEnum && Enum.GetUnderlyingType (argType) == delArgType)
+					match = true;
 			}
 
 			return match;
@@ -223,7 +229,7 @@ namespace System
 			}
 			if (!argLengthMatch) {
 				if (throwOnBindFailure)
-					throw new ArgumentException ("method argument length mismatch");
+					throw new TargetParameterCountException ("Parameter count mismatch.");
 				else
 					return null;
 			}
@@ -495,8 +501,11 @@ namespace System
 
 		public override int GetHashCode ()
 		{
-			/* same implementation as CoreCLR */
-			return GetType ().GetHashCode ();
+			MethodInfo m;
+
+			m = Method;
+
+			return (m != null ? m.GetHashCode () : GetType ().GetHashCode ()) ^ RuntimeHelpers.GetHashCode (m_target);
 		}
 
 		protected virtual MethodInfo GetMethodImpl ()
@@ -534,6 +543,12 @@ namespace System
 			if (a == null)
 				return b;
 
+			if (b == null)
+				return a;
+
+			if (a.GetType () != b.GetType ())
+				throw new ArgumentException (Locale.GetText ("Incompatible Delegate Types. First is {0} second is {1}.", a.GetType ().FullName, b.GetType ().FullName));
+
 			return a.CombineImpl (b);
 		}
 
