@@ -337,7 +337,9 @@ mono_de_add_pending_breakpoints (MonoMethod *method, MonoJitInfo *ji)
 				}
 			}
 
-			g_assert (seq_points);
+			if (!seq_points)
+				/* Could be AOT code, or above "search_all_backends" call could have failed */
+				continue;
 
 			insert_breakpoint (seq_points, domain, ji, bp, NULL);
 		}
@@ -442,12 +444,12 @@ mono_de_set_breakpoint (MonoMethod *method, long il_offset, EventRequest *req, M
 
 	mono_loader_lock ();
 
-	CollectDomainData user_data = {
-		.bp = bp,
-		.methods = methods,
-		.method_domains = method_domains,
-		.method_seq_points = method_seq_points
-	};
+	CollectDomainData user_data;
+	memset (&user_data, 0, sizeof (user_data));
+	user_data.bp = bp;
+	user_data.methods = methods;
+	user_data.method_domains = method_domains;
+	user_data.method_seq_points = method_seq_points;
 	mono_de_foreach_domain (collect_domain_bp, &user_data);
 
 	for (i = 0; i < methods->len; ++i) {
@@ -473,6 +475,17 @@ mono_de_set_breakpoint (MonoMethod *method, long il_offset, EventRequest *req, M
 	return bp;
 }
 
+MonoBreakpoint *
+mono_de_get_breakpoint_by_id (int id)
+{
+	for (int i = 0; i < breakpoints->len; ++i) {
+		MonoBreakpoint *bp = (MonoBreakpoint *)g_ptr_array_index (breakpoints, i);
+		if (bp->req->id == id)
+			return bp;
+	}
+	return NULL;
+}
+
 void
 mono_de_clear_breakpoint (MonoBreakpoint *bp)
 {
@@ -920,7 +933,7 @@ mono_de_ss_update (SingleStepReq *req, MonoJitInfo *ji, SeqPoint *sp, void *tls,
 		return FALSE;
 	}
 
-	if (req->depth == STEP_DEPTH_OVER && (sp->flags & MONO_SEQ_POINT_FLAG_NONEMPTY_STACK)) {
+	if (req->depth == STEP_DEPTH_OVER && (sp->flags & MONO_SEQ_POINT_FLAG_NONEMPTY_STACK) && !(sp->flags & MONO_SEQ_POINT_FLAG_NESTED_CALL)) {
 		/*
 		 * These seq points are inserted by the JIT after calls, step over needs to skip them.
 		 */
@@ -1002,7 +1015,6 @@ mono_de_process_breakpoint (void *void_tls, gboolean from_signal)
 	guint8 *ip;
 	int i;
 	guint32 native_offset;
-	MonoBreakpoint *bp;
 	GPtrArray *bp_reqs, *ss_reqs_orig, *ss_reqs;
 	EventKind kind = EVENT_KIND_BREAKPOINT;
 	MonoContext *ctx = rt_callbacks.tls_get_restore_state (tls);
@@ -1050,7 +1062,6 @@ mono_de_process_breakpoint (void *void_tls, gboolean from_signal)
 
 	mono_debugger_log_bp_hit (tls, method, sp.il_offset);
 
-	bp = NULL;
 	mono_de_collect_breakpoints_by_sp (&sp, ji, ss_reqs_orig, bp_reqs);
 
 	if (bp_reqs->len == 0 && ss_reqs_orig->len == 0) {
@@ -1140,7 +1151,7 @@ static gboolean
 ss_bp_is_unique (GSList *bps, GHashTable *ss_req_bp_cache, MonoMethod *method, guint32 il_offset)
 {
 	if (ss_req_bp_cache) {
-		MonoBreakpoint dummy = {method, il_offset, NULL, NULL};
+		MonoBreakpoint dummy = {method, (long)il_offset, NULL, NULL};
 		return !g_hash_table_lookup (ss_req_bp_cache, &dummy);
 	}
 	for (GSList *l = bps; l; l = l->next) {
@@ -1218,7 +1229,7 @@ is_last_non_empty (SeqPoint* sp, MonoSeqPointInfo *info)
 	SeqPoint* next = g_new (SeqPoint, sp->next_len);
 	mono_seq_point_init_next (info, *sp, next);
 	for (int i = 0; i < sp->next_len; i++) {
-		if (next [i].flags & MONO_SEQ_POINT_FLAG_NONEMPTY_STACK) {
+		if (next [i].flags & MONO_SEQ_POINT_FLAG_NONEMPTY_STACK && !(next [i].flags & MONO_SEQ_POINT_FLAG_NESTED_CALL)) {
 			if (!is_last_non_empty (&next [i], info)) {
 				g_free (next);
 				return FALSE;
