@@ -1026,7 +1026,7 @@ public class Tests
 		return t.ToString ();
 	}
 
-	enum AnEnum {
+	public enum AnEnum {
 		One,
 		Two
 	};
@@ -1106,10 +1106,17 @@ public class Tests
 
 	interface IConstrainedCalls {
 		Pair<int, int> vtype_ret<T, T2>(T t, T2 t2) where T: IReturnVType;
+		AnEnum enum_ret<T, T2>(T t, T2 t2) where T: IReturnVType;
+		int normal_args<T, T2> (T t, T2 t2, int i1, int i2, string s, ref int i3) where T : IConstrained2;
 	}
 
 	public interface IReturnVType {
 		Pair<int, int> return_vtype ();
+		AnEnum return_enum ();
+	}
+
+	public interface IConstrained2 {
+		int normal_args (int i1, int i2, string s, ref int i3);
 	}
 
 	public class CConstrainedCalls : IConstrainedCalls {
@@ -1117,12 +1124,31 @@ public class Tests
 		public Pair<int, int> vtype_ret<T, T2>(T t, T2 t2) where T : IReturnVType {
 			return t.return_vtype ();
 		}
+
+		[MethodImplAttribute (MethodImplOptions.NoInlining)]
+		public AnEnum enum_ret<T, T2>(T t, T2 t2) where T : IReturnVType {
+			return t.return_enum ();
+		}
+
+		public int normal_args<T, T2> (T t, T2 t2, int i1, int i2, string s, ref int i3) where T : IConstrained2 {
+			return t.normal_args (i1, i2, s, ref i3);
+		}
 	}
 
 	class ReturnVType : IReturnVType {
 		public Pair<int, int> return_vtype () {
 			return new Pair<int, int> () { First = 1, Second = 2 };
 		}
+		public AnEnum return_enum () {
+			return AnEnum.Two;
+		}
+	}
+
+	class ConstrainedCalls : IConstrained2 {
+		public int normal_args (int i1, int i2, string s, ref int i3) {
+			i3 = i3 + 1;
+			return i1 + i2 + i3 + s.Length;
+		}
 	}
 
 	public static int test_0_constrained_vtype_ret () {
@@ -1133,6 +1159,22 @@ public class Tests
 		return 0;
 	}
 
+	public static int test_0_constrained_enum_ret () {
+		IConstrainedCalls c = new CConstrainedCalls ();
+		var r = c.enum_ret<ReturnVType, int> (new ReturnVType (), 1);
+		if (r != AnEnum.Two)
+			return 1;
+		return 0;
+	}
+
+	public static int test_14_constrained_normal_args () {
+		IConstrainedCalls c = new CConstrainedCalls ();
+
+		int val = 3;
+		var r = c.normal_args<ConstrainedCalls, int> (new ConstrainedCalls (), 0, 1, 2, "ABC", ref val);
+		return r + val;
+	}
+
 	public struct Pair<T1, T2> {
 		public T1 First;
 		public T2 Second;
@@ -1308,6 +1350,8 @@ public class Tests
 
 	// FIXME: The runtime-invoke wrapper used by BeginInvoke is not found
 	[Category ("!FULLAOT")]
+	// FIXME: Wasm is single threaded and can't handle blocking waits
+	[Category ("!WASM")]
 	public static int test_0_begin_end_invoke () {
 		IFace6 o = new Class6 ();
 		var arr1 = o.Del (1);
@@ -1659,6 +1703,10 @@ public class Tests
 				   uint i1, uint i2, uint i3, uint i4);
 		int Structs (T t, int dummy1, int a2, int a3, int a4, int a5, int a6, int a7, int dummy8,
 					 BStruct s);
+		int Floats (T t, double d1, double d2, double d3, double d4, double d5, double d6, double d7, double d8,
+					double d9, double d10, float s11, float s12);
+		void Generic<T2> (T t, T2[] arr, int dummy1, int a2, int a3, int a4, int a5, int a6, int a7, int dummy8,
+						  T2 i1, T2 i2, T2 i3, T2 i4);
 	}
 
 	class Foo3<T> : IFoo3<T> {
@@ -1690,6 +1738,16 @@ public class Tests
 							BStruct s) {
 			return s.a + s.b + s.c + s.d;
 		}
+		public int Floats (T t, double d1, double d2, double d3, double d4, double d5, double d6, double d7, double d8,
+						   double d9, double d10, float s11, float s12) {
+			return (int)d9 + (int)d10 + (int)s11 + (int)s12;
+		}
+		public void Generic<T2> (T t, T2[] arr, int a1, int a2, int a3, int a4, int a5, int a6, int a7, int a8, T2 i1, T2 i2, T2 i3, T2 i4) {
+			arr [0] = i1;
+			arr [1] = i2;
+			arr [2] = i3;
+			arr [3] = i4;
+		}
 	}
 
 	// Passing small normal arguments on the stack
@@ -1713,6 +1771,13 @@ public class Tests
 		int res6 = o.UInts (new EmptyStruct (), 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4);
 		if (res6 != 10)
 			return 6;
+		int res7 = o.Floats (new EmptyStruct (), 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 10.0, 20.0, 30.0f, 40.0f);
+		if (res7 != 100)
+			return 7;
+		int[] arr = new int [4];
+		o.Generic<int> (new EmptyStruct (), arr, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4);
+		if (arr [0] != 1 || arr [1] != 2 || arr [2] != 3 || arr [3] != 4)
+			return 8;
 		return 0;
 	}
 
@@ -1792,6 +1857,19 @@ public class Tests
 		}
 	}
 
+	struct StructTest : IFaceTest {
+
+		int i;
+
+		public StructTest (int arg) {
+			i = arg;
+		}
+
+		public int iface_method () {
+			return i;
+		}
+	}
+
 	// Test constrained calls on an interface made from gsharedvt methods
 	public static int test_42_gsharedvt_constrained_iface () {
 		IFaceConstrainedIFace obj = new ConstrainedIFace ();
@@ -1799,6 +1877,12 @@ public class Tests
 		return obj.foo<IFaceTest, int> (ref t);
 	}
 
+	public static int test_42_gsharedvt_constrained_iface_vtype () {
+		IFaceConstrainedIFace obj = new ConstrainedIFace ();
+		IFaceTest t = new StructTest (42);
+		return obj.foo<IFaceTest, int> (ref t);
+	}
+
 	// Sign extension tests
 	// 0x55   == 85    == 01010101
 	// 0xAA   == 170   == 10101010
@@ -1941,6 +2025,49 @@ public class Tests
 		gsharedvt_vphi (0);
 		return 0;
 	}
+
+	struct AStruct3<T1, T2, T3> {
+		T1 t1;
+		T2 t2;
+		T3 t3;
+	}
+
+	interface IFaceIsRef {
+		bool is_ref<T> ();
+	}
+
+	class ClassIsRef : IFaceIsRef {
+		[MethodImplAttribute (MethodImplOptions.NoInlining)]
+		public bool is_ref<T> () {
+			return RuntimeHelpers.IsReferenceOrContainsReferences<T> ();
+		}
+	}
+
+	public static int test_0_isreference_intrins () {
+		IFaceIsRef iface = new ClassIsRef ();
+		if (iface.is_ref<AStruct3<int, int, int>> ())
+			return 1;
+		if (!iface.is_ref<AStruct3<string, int, int>> ())
+			return 2;
+		return 0;
+	}
+
+	interface IFace59956 {
+		int foo<T> ();
+	}
+
+	class Impl59956 : IFace59956 {
+		public int foo<T> () {
+			var builder = new SparseArrayBuilder<T>(true);
+
+			return builder.Markers._count;
+		}
+	}
+
+	public static int test_1_59956_regress () {
+		IFace59956 iface = new Impl59956 ();
+		return iface.foo<int> ();
+	}
 }
 
 // #13191
@@ -1958,6 +2085,35 @@ public class MobileServiceCollection<TTable, TCol>
 	}
 }
 
+// #59956
+internal struct Marker
+{
+	public Marker(int count, int index) {
+	}
+}
+
+public struct ArrayBuilder<T>
+{
+	private T[] _array;
+	public int _count;
+
+	public ArrayBuilder(int capacity) {
+		_array = new T[capacity];
+		_count = 1;
+	}
+}
+
+internal struct SparseArrayBuilder<T>
+{
+	private ArrayBuilder<Marker> _markers;
+
+	public SparseArrayBuilder(bool initialize) : this () {
+		_markers = new ArrayBuilder<Marker> (10);
+	}
+
+	public ArrayBuilder<Marker> Markers => _markers;
+}
+
 #if !__MOBILE__
 public class GSharedTests : Tests {
 }
