@@ -1,6 +1,7 @@
 using System;
 using System.Text;
 using System.Reflection;
+using System.Collections.Generic;
 using System.Runtime.InteropServices;
 using System.Runtime.CompilerServices;
 
@@ -26,6 +27,11 @@ using System.Runtime.CompilerServices;
  * the IL code looks.
  */
 
+#if __MOBILE__
+namespace ObjectTests
+{
+#endif
+
 struct Simple {
 	public int a;
 	public byte b;
@@ -114,9 +120,11 @@ struct Gamma {
 
 class Tests {
 
-	static int Main () {
-		return TestDriver.RunTests (typeof (Tests));
+#if !__MOBILE__
+	public static int Main (string[] args) {
+		return TestDriver.RunTests (typeof (Tests), args);
 	}
+#endif
 	
 	public static int test_0_return () {
 		Simple s;
@@ -632,6 +640,7 @@ class Tests {
 	public static int test_0_vector_array_cast () {
 		Array arr1 = Array.CreateInstance (typeof (int), new int[] {1}, new int[] {0});
 		Array arr2 = Array.CreateInstance (typeof (int), new int[] {1}, new int[] {10});
+		Array arr5 = Array.CreateInstance (typeof (string), new int[] {1}, new int[] {10});
 
 		if (arr1.GetType () != typeof (int[]))
 			return 1;
@@ -652,6 +661,9 @@ class Tests {
 
 		if (arr2 is int[])
 			return 4;
+		var as_object_arr = arr5 as object [];
+		if (as_object_arr != null)
+			return 5;
 
 		int [,] [] arr3 = new int [1, 1] [];
 		object o = arr3;
@@ -776,6 +788,11 @@ class Tests {
 		}
 		if (!ok)
 			return 12;
+
+		object arr = new int [10];
+		if (arr is IList<int?>)
+			return 13;
+
 		return 0;
 	}
 
@@ -867,6 +884,49 @@ class Tests {
 		return 2;
 	}
 
+	class InstanceDelegateTest {
+		public int a;
+
+		public int return_field () {
+			return a;
+		}
+	}
+
+	public static int test_2_instance_delegate_with_field () {
+		InstanceDelegateTest t = new InstanceDelegateTest () { a = 1337 };
+		GetIntDel del = new GetIntDel (t.return_field);
+		int v = del ();
+		if (v != 1337)
+			return 0;
+		return 2;
+	}
+
+	interface IFaceVirtualDel {
+		int return_field ();
+	}
+
+	struct VtypeVirtualDelStruct : IFaceVirtualDel {
+		public int f;
+		public int return_field_nonvirt () {
+			return f;
+		}
+		public int return_field () {
+			return f;
+		}
+	}
+
+	public static int test_42_vtype_delegate () {
+		var s = new VtypeVirtualDelStruct () { f = 42 };
+		Func<int> f = s.return_field_nonvirt;
+		return f ();
+	}
+
+	public static int test_42_vtype_virtual_delegate () {
+		IFaceVirtualDel s = new VtypeVirtualDelStruct () { f = 42 };
+		Func<int> f = s.return_field;
+		return f ();
+	}
+
 	public static int test_1_store_decimal () {
 		decimal[,] a = {{1}};
 
@@ -981,6 +1041,19 @@ class Tests {
 		return 0;
 	}
 
+	static Nullable<bool> s_nullb;
+	static AStruct s_struct1;
+
+	/* test if VES uses correct sizes for value type write to static field */
+	public static int test_0_static_nullable_bool () {
+		s_struct1 = new AStruct (0x1337dead);
+		s_nullb = true;
+		/* make sure that the write to s_nullb didn't smash the value after it */
+		if (s_struct1.i != 0x1337dead)
+			return 2;
+		return 0;
+	}
+
 	public static int test_71_long_shift_right () {
 		ulong value = 38654838087;
 		int x = 0;
@@ -1078,7 +1151,8 @@ class Tests {
 	}
 	public static int test_0_cond_branch_side_effects () {
 		counter = 5;
-		if (WriteStuff());
+		if (WriteStuff()) {
+		}
 		if (counter == 10)
 			return 0;
 		return 1;
@@ -1198,6 +1272,27 @@ ncells ) {
 		return sb.ToString () == "ADC" ? 0 : 1;
 	}
 
+	enum FlagsEnum {
+		None = 0,
+		A = 1,
+		B = 2,
+		C = 4
+	}
+
+	public static int test_0_intrins_enum_hasflag () {
+		var e = FlagsEnum.A | FlagsEnum.B;
+
+		if (!e.HasFlag (FlagsEnum.A))
+			return 1;
+		if (!e.HasFlag (FlagsEnum.A | FlagsEnum.B))
+			return 2;
+		if (!e.HasFlag (FlagsEnum.None))
+			return 3;
+		if (e.HasFlag (FlagsEnum.C))
+			return 4;
+		return 0;
+	}
+
 	public class Bar {
 		bool allowLocation = true;
         Foo f = new Foo ();	
@@ -1294,6 +1389,8 @@ ncells ) {
 		return 1.4e-45f;
 	}
 
+#if !NO_BITCODE
+	[Category ("!BITCODE")] // bug #59953
 	public static int test_0_float_return_spill () {
 		// The return value of return_float () is spilled because of the
 		// boxing call
@@ -1301,6 +1398,7 @@ ncells ) {
 		float f = return_float ();
 		return (float)o == f ? 0 : 1;
 	}
+#endif
 
 	class R4Holder {
 		public static float pi = 3.14f;
@@ -1520,9 +1618,391 @@ ncells ) {
 		return 0;
 	}
 
+	/*
+	 * FIXME: This fails with AOT #703317.
+	 */
+	/*
 	static int test_0_multiple_cctor_calls_regress_679467 () {
 		flag = false;
 		return regress_679467_inner ();
 	}
+	*/
+
+	static int test_0_char_ctor () {
+		string s = new String (new char[] { 'A', 'B' }, 0, 1);
+		return 0;
+	}
+
+	static object mInstance = null;
+
+	[MethodImpl(MethodImplOptions.Synchronized)]
+	public static object getInstance() {
+		if (mInstance == null)
+			mInstance = new object();
+		return mInstance;
+	}
+
+	static int test_0_synchronized () {
+		getInstance ();
+		return 0;
+	}
+
+	struct BStruct {
+		public Type t;
+	}
+
+	class Del<T> {
+		public static BStruct foo () {
+			return new BStruct () { t = typeof (T) };
+		}
+	}
+
+	delegate BStruct ADelegate ();
+
+	static int test_0_regress_10601 () {
+		var act = (ADelegate)(Del<string>.foo);
+		BStruct b = act ();
+		if (b.t != typeof (string))
+			return 1;
+		return 0;
+	}
+
+	static int test_0_regress_11058 () {
+		int foo = -252674008;
+		int foo2 = (int)(foo ^ 0xF0F0F0F0); // = 28888
+		var arr = new byte[foo2].Length;
+		return 0;
+	}
+
+	public static void do_throw () {
+		throw new Exception ();
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	static void empty () {
+	}
+
+	// #11297
+	public static int test_0_llvm_inline_throw () {
+		try {
+			empty ();
+		} catch (Exception) {
+			do_throw ();
+		}
+
+		return 0;
+	}
+
+	enum ByteEnum : byte {
+        Zero = 0
+    }
+
+    struct BugStruct {
+        public ByteEnum f1;
+        public ByteEnum f2;
+        public ByteEnum f3;
+        public byte f4;
+        public byte f5;
+        public byte f6;
+        public byte f7;
+    }
+
+	public static int test_0_14217 () {
+		t_14217_inner (new BugStruct ());
+		return 0;
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	static void t_14217_inner (BugStruct bug) {
+    }
+
+	[StructLayout(LayoutKind.Sequential)]
+	public struct EmptyStruct {
+	}
+
+	class EmptyClass {
+		public static EmptyStruct s;
+	}
+
+	// #20349
+	static int test_0_empty_struct_as_static () {
+		var s = EmptyClass.s;
+		return 0;
+	}
+
+	// #25487
+	static int test_0_int_to_r4 () {
+		return int_to_r4_inner (255);
+	}
+
+	static int int_to_r4_inner (int value1) {
+		int sub = -value1;
+		float mult = sub * 1f;
+		if (mult != -255.0f)
+			return 1;
+		else
+			return 0;
+	}
+
+	struct HFA4D {
+		public double a, b, c, d;
+	}
+
+	static double arm64_hfa_on_stack_inner (double d1, double d2, double d3, double d4, double d5, double d6, double d7, double d8, HFA4D s) {
+		return s.a + s.b + s.c + s.d;
+	}
+
+	static int test_0_arm64_hfa_on_stack () {
+		var s = new HFA4D () { a = 1.0, b = 2.0, c = 3.0, d = 4.0 };
+		var res = arm64_hfa_on_stack_inner (1, 2, 3, 4, 5, 6, 7, 8, s);
+		return res == 10.0 ? 0 : 1;
+	}
+
+	class MulOvfClass {
+		[MethodImplAttribute (MethodImplOptions.NoInlining)]
+		public unsafe void EncodeIntoBuffer(char* value, int valueLength, char* buffer, int bufferLength) {
+		}
+	}
+
+	static unsafe int test_0_mul_ovf_regress_36052 () {
+		var p = new MulOvfClass ();
+
+		string typeName = typeof(int).Name;
+		int bufferSize = 45;
+
+		fixed (char* value = typeName) {
+			char* buffer = stackalloc char[bufferSize];
+			p.EncodeIntoBuffer(value, typeName.Length, buffer, bufferSize);
+		}
+		return 0;
+	}
+
+	struct Struct16 {
+		public int a, b, c, d;
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	static int pass_struct16 (object o0, object o2, object o3, object o4, object o5, object o6, object o7, Struct16 o8) {
+		// This disables LLVM
+		try {
+		} catch {
+		}
+		return o8.a;
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	static int pass_struct16 (object o0, object o2, object o3, object o6, object o7, Struct16 o8) {
+		return pass_struct16 (o0, o2, null, o3, null, o6, o7, o8);
+	}
+
+	public static int test_42_pass_16byte_struct_split () {
+		return pass_struct16 (null, null, null, null, null, new Struct16 () { a = 42 });
+	}
+
+	public interface IComparer2
+	{
+		Type foo<T> ();
+	}
+
+	public class AClass : IComparer2 {
+		public Type foo<T> () {
+			return typeof(T);
+		}
+	}
+
+	public static int test_0_delegate_to_virtual_generic_on_ifaces () {
+		IComparer2 c = new AClass ();
+
+		Func<Type> f = c.foo<string>;
+		return f () == typeof(string) ? 0 : 1;
+	}
+
+	public enum ByteEnum2 : byte {
+		High = 142
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static int enum_arg_zero_extend (ByteEnum2 b) {
+		return (int)b;
+	}
+
+	public static int test_142_byte_enum_arg_zero_extend () {
+		return enum_arg_zero_extend (ByteEnum2.High);
+	}
+
+	enum Mine { One, Two }
+
+	public static int test_0_enum_gethashcode_opt () {
+		int sum = 0;
+        for (int i = 0; i < 1000000; ++i)
+			sum += Mine.Two.GetHashCode();
+
+        return 0;
+    }
+
+	public static int test_0_typedref () {
+		int i = 5;
+		System.TypedReference r = __makeref(i);
+		System.Type t = __reftype(r);
+
+		if (t != typeof (int))
+			return 1;
+		int j = __refvalue(r, int);
+		if (j != 5)
+			return 2;
+
+		try {
+			object o = __refvalue (r, object);
+		} catch (InvalidCastException) {
+		}
+
+		return 0;
+	}
+
+	enum FooEnum { Bar }
+	//https://github.com/mono/mono/issues/6666
+	public static int test_0_bad_unbox_nullable_of_enum () {
+		try {
+			var enumValue = FooEnum.Bar;
+			object value = (int)enumValue;
+			var res = (FooEnum?)value; // Should throw
+		} catch (InvalidCastException) {
+			return 0;
+		}
+		return 1;
+	}
+
+	//https://github.com/mono/mono/issues/6666
+	public static int test_0_unbox_nullable_of_enum () {
+		try {
+			var enumValue = FooEnum.Bar;
+			object value = (object)enumValue;
+			var res = (FooEnum?)value; // Should not throw
+		} catch (InvalidCastException) {
+			return 1;
+		}
+		return 0;
+	}
+
+	static void decode (out sbyte v) {
+		byte tmp = 134;
+		v = (sbyte)tmp;
+	}
+
+	// gh #6414
+	public static int test_0_alias_analysis_sign_extend () {
+	  sbyte t;
+	  decode (out t);
+
+	  return t == -122 ? 0 : 1;
+	}
+
+	public interface IFoo
+	{
+	  int MyInt { get; }
+	}
+
+	public class IFooImpl : IFoo
+	{
+	  public int MyInt => 0;
+	}
+
+	//gh 6266
+    public static int test_0_store_to_magic_iface_array ()
+    {
+      ICollection<IFoo> arr1 = new IFooImpl[1] { new IFooImpl() };
+      ICollection<IFoo> arr2 = new IFooImpl[1] { new IFooImpl() };
+
+      ICollection<IFoo>[] a2d = new ICollection<IFoo>[2] {
+        arr1,
+        arr2,
+      };
+
+	  return 0;
+    }
+
+	static volatile bool abool;
+
+	public static unsafe int test_0_stind_r4_float32_stack_merge () {
+		Single* dataPtr = stackalloc Single[4];
+		abool = true;
+		dataPtr[0] = abool ? 1.0f : 2.0f;
+		return dataPtr [0] == 1.0f ? 0 : 1;
+	}
+
+	class AClass1 {
+	}
+
+	class BClass1 : AClass1 {
+	}
+
+	class CClass1 {
+	}
+
+	public static int test_0_array_of_magic_iface () {
+		// Need to make this an object otherwise csc removes the cast
+		object d = new [] { new [] { new BClass1 () } };
+		if (!(d is IList<AClass1> []))
+			return 1;
+		if (d is IList<CClass1> [])
+			return 2;
+		var e2 = (IList<AClass1> []) d;
+		return 0;
+	}
+
+	class SimpleContainer {
+		public Simple simple1;
+		public Simple simple2;
+
+		public static Simple constsimple;
+
+		public int SetFields () {
+			constsimple.a = 0x1337;
+			simple1 = simple2 = constsimple;
+			return simple1.a - simple2.a;
+		}
+	}
+
+	public static int test_0_dup_vtype () {
+		return new SimpleContainer ().SetFields ();
+	}
+
+	public struct Vec3 {
+		public int X, Y, Z;
+
+		[MethodImplAttribute (MethodImplOptions.NoInlining)]
+			public Vec3(int x, int y, int z) {
+			X = x;
+			Y = y;
+			Z = z;
+		}
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static int gh_11378_inner_1 (Vec3 p1, Vec3 p2) {
+		p1.X -= p2.X;
+		p1.Y -= p2.Y;
+		p1.Z -= p2.Z;
+
+		return (int)p2.Y;
+	}
+
+	[MethodImplAttribute (MethodImplOptions.NoInlining)]
+	public static int gh_11378_inner_2 (Vec3 c, Vec3 pos) {
+		return gh_11378_inner_1 (pos, c);
+	}
+
+	static int gh_11378_inner_3 (Vec3 c) {
+		var c2 = c;
+		return gh_11378_inner_2 (c, c2);
+	}
+
+	public static int test_2_gh_11378 () {
+		return gh_11378_inner_3 (new Vec3(0, 2, -20));
+	}
+
 }
 
+#if __MOBILE__
+}
+#endif
