@@ -1,21 +1,25 @@
-/*
- * ssa.c: Static single assign form support for the JIT compiler.
+/**
+ * \file
+ * Static single assign form support for the JIT compiler.
  *
  * Author:
  *    Dietmar Maurer (dietmar@ximian.com)
  *
  * (C) 2003 Ximian, Inc.
  * Copyright 2011 Xamarin, Inc (http://www.xamarin.com)
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 #include <config.h>
 #include <string.h>
 #include <mono/metadata/debug-helpers.h>
 #include <mono/metadata/mempool.h>
 #include <mono/metadata/mempool-internals.h>
+#include <mono/utils/mono-compiler.h>
 
 #ifndef DISABLE_JIT
 
 #include "mini.h"
+#include "mini-runtime.h"
 #ifdef HAVE_ALLOCA_H
 #include <alloca.h>
 #endif
@@ -157,9 +161,8 @@ typedef struct {
 
 /**
  * mono_ssa_rename_vars:
- *
- *  Implement renaming of SSA variables. Also compute def-use information in parallel.
- * @stack_history points to an area of memory which can be used for storing changes 
+ * Implement renaming of SSA variables. Also compute def-use information in parallel.
+ * \p stack_history points to an area of memory which can be used for storing changes 
  * made to the stack, so they can be reverted later.
  */
 static void
@@ -236,8 +239,13 @@ mono_ssa_rename_vars (MonoCompile *cfg, int max_vars, MonoBasicBlock *bb, gboole
 				if (var->opcode == OP_ARG)
 					originals_used [idx] = TRUE;
 
-				/* FIXME: */
-				g_assert (stack_history_len < stack_history_size);
+				if (stack_history_len + 128 > stack_history_size) {
+					stack_history_size += 1024;
+					RenameInfo *new_history = mono_mempool_alloc (cfg->mempool, sizeof (RenameInfo) * stack_history_size);
+					memcpy (new_history, stack_history, stack_history_len * sizeof (RenameInfo));
+					stack_history = new_history;
+				}
+
 				stack_history [stack_history_len].var = stack [idx];
 				stack_history [stack_history_len].idx = idx;
 				stack_history_len ++;
@@ -457,7 +465,7 @@ mono_ssa_compute (MonoCompile *cfg)
 
 	/* Renaming phase */
 
-	stack = (MonoInst **)alloca (sizeof (MonoInst *) * cfg->num_varinfo);
+	stack = g_newa (MonoInst*, cfg->num_varinfo);
 	memset (stack, 0, sizeof (MonoInst *) * cfg->num_varinfo);
 
 	lvreg_stack = g_new0 (guint32, cfg->next_vreg);
@@ -501,7 +509,7 @@ mono_ssa_remove_gsharedvt (MonoCompile *cfg)
 			printf ("\nREMOVE SSA %d:\n", bb->block_num);
 
 		for (ins = bb->code; ins; ins = ins->next) {
-			if (!(MONO_IS_PHI (ins) && ins->opcode == OP_VPHI && mini_is_gsharedvt_variable_type (&ins->klass->byval_arg)))
+			if (!(MONO_IS_PHI (ins) && ins->opcode == OP_VPHI && mini_is_gsharedvt_variable_type (m_class_get_byval_arg (ins->klass))))
 				continue;
 
 			g_assert (ins->inst_phi_args [0] == bb->in_count);
@@ -1501,7 +1509,7 @@ mono_ssa_loop_invariant_code_motion (MonoCompile *cfg)
 				MONO_REMOVE_INS (bb, ins);
 				mono_bblock_insert_before_ins (idom, idom->last_ins, ins);
 				if (ins->opcode == OP_LDLEN || ins->opcode == OP_STRLEN)
-					idom->has_array_access = TRUE;
+					idom->needs_decompose = TRUE;
 			}
 		}
 	}
@@ -1514,4 +1522,8 @@ mono_ssa_loop_invariant_code_motion (MonoCompile *cfg)
 	}
 }
 
-#endif /* DISABLE_JIT */
+#else /* !DISABLE_JIT */
+
+MONO_EMPTY_SOURCE_FILE (ssa);
+
+#endif /* !DISABLE_JIT */
