@@ -1,5 +1,6 @@
-/*
- * trace.c: Tracing facilities for the Mono Runtime.
+/**
+ * \file
+ * Tracing facilities for the Mono Runtime.
  *
  * Author:
  *   Paolo Molaro (lupus@ximian.com)
@@ -7,6 +8,7 @@
  *
  * (C) 2002 Ximian, Inc.
  * Copyright 2011 Xamarin, Inc (http://www.xamarin.com)
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include <config.h>
@@ -18,333 +20,49 @@
 #endif
 #include <string.h>
 #include "mini.h"
+#include "mini-runtime.h"
 #include <mono/metadata/debug-helpers.h>
-#include <mono/metadata/assembly.h>
 #include <mono/utils/mono-time.h>
 #include <mono/utils/mono-memory-model.h>
 #include "trace.h"
+#include <mono/metadata/callspec.h>
 
-#if defined (PLATFORM_ANDROID) || (defined (TARGET_IOS) && defined (TARGET_IOS))
+#if defined (HOST_ANDROID) || (defined (TARGET_IOS) && defined (TARGET_IOS))
 #  undef printf
 #  define printf(...) g_log("mono", G_LOG_LEVEL_MESSAGE, __VA_ARGS__)
 #  undef fprintf
 #  define fprintf(__ignore, ...) g_log ("mono-gc", G_LOG_LEVEL_MESSAGE, __VA_ARGS__)
 #endif
 
-#ifdef __GNUC__
-
-#define RETURN_ADDRESS_N(N) (__builtin_extract_return_addr (__builtin_return_address (N)))
-#define RETURN_ADDRESS() RETURN_ADDRESS_N(0)
-
-#elif defined(_MSC_VER)
-
-#include <intrin.h>
-#pragma intrinsic(_ReturnAddress)
-
-#define RETURN_ADDRESS() _ReturnAddress()
-#define RETURN_ADDRESS_N(N) NULL
-
-#else
-
-#error "Missing return address intrinsics implementation"
-
-#endif
-
-static MonoTraceSpec trace_spec;
+static MonoCallSpec trace_spec;
 
 static volatile gint32 output_lock = 0;
 
-gboolean
-mono_trace_eval_exception (MonoClass *klass)
+gboolean mono_trace_eval_exception (MonoClass *klass)
 {
-	int include = 0;
-	int i;
-
-	if (!klass)
-		return FALSE;
-
-	for (i = 0; i < trace_spec.len; i++) {
-		MonoTraceOperation *op = &trace_spec.ops [i];
-		int inc = 0;
-		
-		switch (op->op){
-		case MONO_TRACEOP_EXCEPTION:
-			if (strcmp ("", op->data) == 0 && strcmp ("all", op->data2) == 0)
-				inc = 1;
-			else if (strcmp ("", op->data) == 0 || strcmp (klass->name_space, op->data) == 0)
-				if (strcmp (klass->name, op->data2) == 0)
-					inc = 1;
-			break;
-		default:
-			break;
-		}
-		if (op->exclude){
-			if (inc)
-				include = 0;
-		} else if (inc)
-			include = 1;
-	}
-
-	return include;
+	return mono_callspec_eval_exception (klass, &trace_spec);
 }
 
-gboolean
-mono_trace_eval (MonoMethod *method)
+gboolean mono_trace_eval (MonoMethod *method)
 {
-	int include = 0;
-	int i;
-
-	for (i = 0; i < trace_spec.len; i++){
-		MonoTraceOperation *op = &trace_spec.ops [i];
-		int inc = 0;
-		
-		switch (op->op){
-		case MONO_TRACEOP_ALL:
-			inc = 1; break;
-		case MONO_TRACEOP_PROGRAM:
-			if (trace_spec.assembly && (method->klass->image == mono_assembly_get_image (trace_spec.assembly)))
-				inc = 1; break;
-		case MONO_TRACEOP_WRAPPER:
-			if ((method->wrapper_type == MONO_WRAPPER_NATIVE_TO_MANAGED) ||
-				(method->wrapper_type == MONO_WRAPPER_MANAGED_TO_NATIVE))
-				inc = 1; break;
-		case MONO_TRACEOP_METHOD:
-			if (mono_method_desc_full_match ((MonoMethodDesc *) op->data, method))
-				inc = 1; break;
-		case MONO_TRACEOP_CLASS:
-			if (strcmp (method->klass->name_space, op->data) == 0)
-				if (strcmp (method->klass->name, op->data2) == 0)
-					inc = 1;
-			break;
-		case MONO_TRACEOP_ASSEMBLY:
-			if (strcmp (mono_image_get_name (method->klass->image), op->data) == 0)
-				inc = 1; break;
-		case MONO_TRACEOP_NAMESPACE:
-			if (strcmp (method->klass->name_space, op->data) == 0)
-				inc = 1;
-		case MONO_TRACEOP_EXCEPTION:
-			break;
-		}
-		if (op->exclude){
-			if (inc)
-				include = 0;
-		} else if (inc)
-			include = 1;
-	}
-	return include;
+	return mono_callspec_eval (method, &trace_spec);
 }
 
-static int is_filenamechar (char p)
+MonoCallSpec *mono_trace_set_options (const char *options)
 {
-	if (p >= 'A' && p <= 'Z')
-		return TRUE;
-	if (p >= 'a' && p <= 'z')
-		return TRUE;
-	if (p >= '0' && p <= '9')
-		return TRUE;
-	if (p == '.' || p == ':' || p == '_' || p == '-' || p == '`')
-		return TRUE;
-	return FALSE;
-}
-
-static char *input;
-static char *value;
-
-static void get_string (void)
-{
-	char *start = input;
-	while (is_filenamechar (*input)){
-		input++;
-	}
-	if (value != NULL)
-		g_free (value);
-	value = (char *)g_malloc (input - start + 1);
-	strncpy (value, start, input-start);
-	value [input-start] = 0;
-}
-
-enum Token {
-	TOKEN_METHOD,
-	TOKEN_CLASS,
-	TOKEN_ALL,
-	TOKEN_PROGRAM,
-	TOKEN_EXCEPTION,
-	TOKEN_NAMESPACE,
-	TOKEN_WRAPPER,
-	TOKEN_STRING,
-	TOKEN_EXCLUDE,
-	TOKEN_DISABLED,
-	TOKEN_SEPARATOR,
-	TOKEN_END,
-	TOKEN_ERROR
-};
-
-static int
-get_token (void)
-{
-	while (input [0] == '+')
-		input++;
-
-	if (input [0] == '\0') {
-		return TOKEN_END;
-	}
-	if (input [0] == 'M' && input [1] == ':'){
-		input += 2;
-		get_string ();
-		return TOKEN_METHOD;
-	}
-	if (input [0] == 'N' && input [1] == ':'){
-		input += 2;
-		get_string ();
-		return TOKEN_NAMESPACE;
-	}
-	if (input [0] == 'T' && input [1] == ':'){
-		input += 2;
-		get_string ();
-		return TOKEN_CLASS;
-	}
-	if (input [0] == 'E' && input [1] == ':'){
-		input += 2;
-		get_string ();
-		return TOKEN_EXCEPTION;
-	}
-	if (*input == '-'){
-		input++;
-		return TOKEN_EXCLUDE;
-	}
-	if (is_filenamechar (*input)){
-		get_string ();
-		if (strcmp (value, "all") == 0)
-			return TOKEN_ALL;
-		if (strcmp (value, "program") == 0)
-			return TOKEN_PROGRAM;
-		if (strcmp (value, "wrapper") == 0)
-			return TOKEN_WRAPPER;
-		if (strcmp (value, "disabled") == 0)
-			return TOKEN_DISABLED;
-		return TOKEN_STRING;
-	}
-	if (*input == ','){
-		input++;
-		return TOKEN_SEPARATOR;
+	char *errstr;
+	if (!mono_callspec_parse (options, &trace_spec, &errstr)) {
+		fprintf (stderr, "%s\n", errstr);
+		g_free (errstr);
+		return NULL;
 	}
 
-	fprintf (stderr, "Syntax error at or around '%s'\n", input);	
-	return TOKEN_ERROR;
-}
-
-static void
-cleanup (void)
-{
-	if (value != NULL)
-		g_free (value);
-}
-
-static int
-get_spec (int *last)
-{
-	int token = get_token ();
-	if (token == TOKEN_EXCLUDE){
-		token = get_spec (last);
-		if (token == TOKEN_EXCLUDE){
-			fprintf (stderr, "Expecting an expression");
-			return TOKEN_ERROR;
-		}
-		if (token == TOKEN_ERROR)
-			return token;
-		trace_spec.ops [(*last)-1].exclude = 1;
-		return TOKEN_SEPARATOR;
-	}
-	if (token == TOKEN_END || token == TOKEN_SEPARATOR || token == TOKEN_ERROR)
-		return token;
-	
-	if (token == TOKEN_METHOD){
-		MonoMethodDesc *desc = mono_method_desc_new (value, TRUE);
-		if (desc == NULL){
-			fprintf (stderr, "Invalid method name: %s\n", value);
-			return TOKEN_ERROR;
-		}
-		trace_spec.ops [*last].op = MONO_TRACEOP_METHOD;
-		trace_spec.ops [*last].data = desc;
-	} else if (token == TOKEN_ALL)
-		trace_spec.ops [*last].op = MONO_TRACEOP_ALL;
-	else if (token == TOKEN_PROGRAM)
-		trace_spec.ops [*last].op = MONO_TRACEOP_PROGRAM;
-	else if (token == TOKEN_WRAPPER)
-		trace_spec.ops [*last].op = MONO_TRACEOP_WRAPPER;
-	else if (token == TOKEN_NAMESPACE){
-		trace_spec.ops [*last].op = MONO_TRACEOP_NAMESPACE;
-		trace_spec.ops [*last].data = g_strdup (value);
-	} else if (token == TOKEN_CLASS || token == TOKEN_EXCEPTION){
-		char *p = strrchr (value, '.');
-		if (p) {
-			*p++ = 0;
-			trace_spec.ops [*last].data = g_strdup (value);
-			trace_spec.ops [*last].data2 = g_strdup (p);
-		}
-		else {
-			trace_spec.ops [*last].data = g_strdup ("");
-			trace_spec.ops [*last].data2 = g_strdup (value);
-		}
-		trace_spec.ops [*last].op = token == TOKEN_CLASS ? MONO_TRACEOP_CLASS : MONO_TRACEOP_EXCEPTION;
-	} else if (token == TOKEN_STRING){
-		trace_spec.ops [*last].op = MONO_TRACEOP_ASSEMBLY;
-		trace_spec.ops [*last].data = g_strdup (value);
-	} else if (token == TOKEN_DISABLED) {
-		trace_spec.enabled = FALSE;
-	} else {
-		fprintf (stderr, "Syntax error in trace option specification\n");
-		return TOKEN_ERROR;
-	}
-	(*last)++;
-	return TOKEN_SEPARATOR;
-}
-
-MonoTraceSpec *
-mono_trace_parse_options (const char *options)
-{
-	char *p = (char*)options;
-	int size = 1;
-	int last_used;
-	int token;
-
-	trace_spec.enabled = TRUE;
-	if (*p == 0){
-		trace_spec.len = 1;
-		trace_spec.ops = g_new0 (MonoTraceOperation, 1);
-		trace_spec.ops [0].op = MONO_TRACEOP_ALL;
-		return &trace_spec;
-	}
-		
-	for (p = (char*)options; *p != 0; p++)
-		if (*p == ',')
-			size++;
-	
-	trace_spec.ops = g_new0 (MonoTraceOperation, size);
-
-	input = (char*)options;
-	last_used = 0;
-	
-	while ((token = (get_spec (&last_used))) != TOKEN_END){
-		if (token == TOKEN_ERROR)
-			return NULL;
-		if (token == TOKEN_SEPARATOR)
-			continue;
-	}
-	trace_spec.len = last_used;
-	cleanup ();
 	return &trace_spec;
 }
 
-void
-mono_trace_set_assembly (MonoAssembly *assembly)
-{
-	trace_spec.assembly = assembly;
-}
-
 static
-#ifdef HAVE_KW_THREAD
-__thread 
+#ifdef MONO_KEYWORD_THREAD
+MONO_KEYWORD_THREAD
 #endif
 int indent_level = 0;
 static guint64 start_time = 0;
@@ -369,17 +87,17 @@ static char *
 string_to_utf8 (MonoString *s)
 {
 	char *as;
-	GError *error = NULL;
+	GError *gerror = NULL;
 
 	g_assert (s);
 
 	if (!s->length)
 		return g_strdup ("");
 
-	as = g_utf16_to_utf8 (mono_string_chars (s), s->length, NULL, NULL, &error);
-	if (error) {
+	as = g_utf16_to_utf8 (mono_string_chars (s), s->length, NULL, NULL, &gerror);
+	if (gerror) {
 		/* Happens with StringBuilders */
-		g_error_free (error);
+		g_error_free (gerror);
 		return g_strdup ("<INVALID UTF8>");
 	}
 	else
@@ -416,7 +134,7 @@ mono_trace_enter_method (MonoMethod *method, char *ebp)
 	if (!trace_spec.enabled)
 		return;
 
-	while (output_lock != 0 || InterlockedCompareExchange (&output_lock, 1, 0) != 0)
+	while (output_lock != 0 || mono_atomic_cas_i32 (&output_lock, 1, 0) != 0)
 		mono_thread_info_yield ();
 
 	fname = mono_method_full_name (method, TRUE);
@@ -425,17 +143,36 @@ mono_trace_enter_method (MonoMethod *method, char *ebp)
 	g_free (fname);
 
 	if (!ebp) {
-		printf (") ip: %p\n", RETURN_ADDRESS_N (1));
+
+// https://github.com/mono/mono/issues/8572
+// Build fails on Redhat 6.8 with GCC 4.4.7.
+// Old gcc does not allow pragma diagnostic within function.
+#if __clang__ || __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
+#define GCC_PRAGMA_DIAGNOSTIC_WITHIN_FUNCTION 1
+#endif
+
+#if GCC_PRAGMA_DIAGNOSTIC_WITHIN_FUNCTION
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunknown-warning-option"
+#pragma GCC diagnostic ignored "-Wframe-address"
+#endif
+
+		printf (") ip: %p\n", MONO_RETURN_ADDRESS_N (1));
+
+#if GCC_PRAGMA_DIAGNOSTIC_WITHIN_FUNCTION
+#pragma GCC diagnostic pop
+#endif
+
 		goto unlock;
 	}
 
 	sig = mono_method_signature (method);
 
-	arg_info = (MonoJitArgumentInfo *)alloca (sizeof (MonoJitArgumentInfo) * (sig->param_count + 1));
+	arg_info = g_newa (MonoJitArgumentInfo, sig->param_count + 1);
 
 	if (method->is_inflated) {
 		/* FIXME: Might be better to pass the ji itself */
-		MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (char *)RETURN_ADDRESS (), NULL);
+		MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (char *)MONO_RETURN_ADDRESS (), NULL);
 		if (ji) {
 			gsctx = mono_jit_info_get_generic_sharing_context (ji);
 			if (gsctx && gsctx->is_gsharedvt) {
@@ -456,7 +193,7 @@ mono_trace_enter_method (MonoMethod *method, char *ebp)
 
 	if (mono_method_signature (method)->hasthis) {
 		gpointer *this_obj = (gpointer *)(ebp + arg_info [0].offset);
-		if (method->klass->valuetype) {
+		if (m_class_is_valuetype (method->klass)) {
 			printf ("value:%p, ", *arg_in_stack_slot(this_obj, gpointer *));
 		} else {
 			o = *arg_in_stack_slot(this_obj, MonoObject *);
@@ -471,7 +208,7 @@ mono_trace_enter_method (MonoMethod *method, char *ebp)
 					printf ("this:[STRING:%p:%s], ", o, as);
 					g_free (as);
 				} else {
-					printf ("this:%p[%s.%s %s], ", o, klass->name_space, klass->name, o->vtable->domain->friendly_name);
+					printf ("this:%p[%s.%s %s], ", o, m_class_get_name_space (klass), m_class_get_name (klass), o->vtable->domain->friendly_name);
 				}
 			} else 
 				printf ("this:NULL, ");
@@ -525,18 +262,19 @@ mono_trace_enter_method (MonoMethod *method, char *ebp)
 			o = *arg_in_stack_slot(cpos, MonoObject *);
 			if (o) {
 				klass = o->vtable->klass;
-		    
+
+				gpointer data = mono_object_get_data (o);
 				if (klass == mono_defaults.string_class) {
 					char *as = string_to_utf8 ((MonoString*)o);
 
 					printf ("[STRING:%p:%s], ", o, as);
 					g_free (as);
 				} else if (klass == mono_defaults.int32_class) {
-					printf ("[INT32:%p:%d], ", o, *(gint32 *)((char *)o + sizeof (MonoObject)));
-				} else if (klass == mono_defaults.monotype_class) {
+					printf ("[INT32:%p:%d], ", o, *(gint32 *)data);
+				} else if (klass == mono_defaults.runtimetype_class) {
 					printf ("[TYPE:%s], ", mono_type_full_name (((MonoReflectionType*)o)->type));
 				} else
-					printf ("[%s.%s:%p], ", klass->name_space, klass->name, o);
+					printf ("[%s.%s:%p], ", m_class_get_name_space (klass), m_class_get_name (klass), o);
 			} else {
 				printf ("%p, ", *arg_in_stack_slot(cpos, gpointer));
 			}
@@ -587,7 +325,7 @@ mono_trace_leave_method (MonoMethod *method, ...)
 	if (!trace_spec.enabled)
 		return;
 
-	while (output_lock != 0 || InterlockedCompareExchange (&output_lock, 1, 0) != 0)
+	while (output_lock != 0 || mono_atomic_cas_i32 (&output_lock, 1, 0) != 0)
 		mono_thread_info_yield ();
 
 	va_start(ap, method);
@@ -599,7 +337,7 @@ mono_trace_leave_method (MonoMethod *method, ...)
 
 	if (method->is_inflated) {
 		/* FIXME: Might be better to pass the ji itself */
-		MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (char *)RETURN_ADDRESS (), NULL);
+		MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (char *)MONO_RETURN_ADDRESS (), NULL);
 		if (ji) {
 			gsctx = mono_jit_info_get_generic_sharing_context (ji);
 			if (gsctx && gsctx->is_gsharedvt) {
@@ -656,14 +394,15 @@ mono_trace_leave_method (MonoMethod *method, ...)
 		MonoObject *o = va_arg (ap, MonoObject *);
 
 		if (o) {
+			gpointer data = mono_object_get_data (o);
 			if (o->vtable->klass == mono_defaults.boolean_class) {
-				printf ("[BOOLEAN:%p:%d]", o, *((guint8 *)o + sizeof (MonoObject)));		
+				printf ("[BOOLEAN:%p:%d]", o, *(guint8 *)data);
 			} else if  (o->vtable->klass == mono_defaults.int32_class) {
-				printf ("[INT32:%p:%d]", o, *((gint32 *)((char *)o + sizeof (MonoObject))));	
+				printf ("[INT32:%p:%d]", o, *(gint32 *)data);
 			} else if  (o->vtable->klass == mono_defaults.int64_class) {
-				printf ("[INT64:%p:%lld]", o, (long long)*((gint64 *)((char *)o + sizeof (MonoObject))));	
+				printf ("[INT64:%p:%lld]", o, (long long)*(gint64 *)data);
 			} else
-				printf ("[%s.%s:%p]", o->vtable->klass->name_space, o->vtable->klass->name, o);
+				printf ("[%s.%s:%p]", m_class_get_name_space (mono_object_class (o)), m_class_get_name (mono_object_class (o)), o);
 		} else
 			printf ("[OBJECT:%p]", o);
 	       
@@ -707,7 +446,7 @@ mono_trace_leave_method (MonoMethod *method, ...)
 		printf ("(unknown return type %x)", mono_method_signature (method)->ret->type);
 	}
 
-	//printf (" ip: %p\n", RETURN_ADDRESS_N (1));
+	//printf (" ip: %p\n", MONO_RETURN_ADDRESS_N (1));
 	printf ("\n");
 	fflush (stdout);
 
